from typing_extensions import Final  # noqa: F401

MAILBOX_CREATE_URL = 'https://api.sendgrid.com/v3/user/webhooks/parse/settings'  # type: Final  # noqa: E501  # yapf: disable
MAILBOX_DETAIL_URL = 'https://api.sendgrid.com/v3/user/webhooks/parse/settings/{}'  # type: Final  # noqa: E501  # yapf: disable

INBOX_URL = 'https://mailserver.lokole.ca/api/email/sendgrid/{}'  # type: Final

MX_RECORD = 'mx.sendgrid.net'  # type: Final



from os import getenv
from pathlib import Path

from flask_wtf import FlaskForm
from requests import post
from wtforms import StringField
from wtforms import SubmitField
from wtforms.validators import ValidationError

from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n
from opwen_email_client.webapp.tasks import register


class RegisterForm(FlaskForm):

    client_name = StringField()
    github_username = StringField()
    github_token = StringField()
    submit = SubmitField()

    def register_client(self):
        path = (Path(getenv('OPWEN_STATE_DIRECTORY', 'lokole/state')) / 'settings.env')
        self._setup_client(str(path))

    def _setup_client(self, path):
        name = self.client_name.data.strip()
        token = self.github_token.data.strip()

        endpoint = AppConfig.EMAIL_SERVER_ENDPOINT or 'mailserver.lokole.ca'
        client_domain = '{}.{}'.format(name, 'lokole.ca')
        client_create_url = 'https://{}/api/email/register/'.format(endpoint)

        response = post(client_create_url,
                        json={'domain': client_domain},
                        headers={'Authorization': 'Bearer {}'.format(token)})
        if response.status_code != 201:
            raise ValidationError(i8n.FAILED_REGISTRATION)
        register.delay(name, token, path)



from typing import Iterable
from typing import Iterator
from typing import Optional

from libcloud.storage.types import ObjectDoesNotExistError
from requests import RequestException
from requests import post as http_post

from opwen_email_server.constants import events
from opwen_email_server.constants import github
from opwen_email_server.services.storage import AzureObjectStorage
from opwen_email_server.utils.log import LogMixin


class BasicAuth(LogMixin):
    def __init__(self, users: dict):
        self._users = dict(users)

    def __call__(self, username, password, required_scopes=None):
        if not username or not password:
            return None

        try:
            user = self._users[username]
        except KeyError:
            self.log_event(events.UNKNOWN_USER, {'username': username})  # noqa: E501  # yapf: disable
            return None

        if user['password'] != password:
            self.log_event(events.BAD_PASSWORD, {'username': username})  # noqa: E501  # yapf: disable
            return None

        scopes = user.get('scopes', [])

        return {'sub': {'name': username, 'scopes': scopes}, 'scope': scopes}


class GithubAuth(LogMixin):
    def __init__(self, organization: str, page_size: int = 50):
        self._organization = organization
        self._page_size = page_size

    def __call__(self, access_token, required_scopes=None):
        if not access_token or not self._organization:
            return None

        try:
            query = self._query_github(access_token)
            login = next(query)
            scopes = list(query)
        except RequestException:
            self.log_event(events.BAD_PASSWORD, {'username': 'access_token'})  # noqa: E501  # yapf: disable
            return None

        return {'sub': {'name': login, 'scopes': scopes}, 'scope': scopes}

    def _query_github(self, access_token: str) -> Iterator[str]:
        cursor = None

        while True:
            response = http_post(
                url=github.GRAPHQL_URL,
                json={
                    'query': '''
                        query($organization:String!, $cursor:String, $first:Int!) {
                            viewer {
                                login
                                organization(login:$organization) {
                                    teams(after:$cursor, first:$first, orderBy:{ field:NAME, direction:DESC }) {
                                        edges {
                                            cursor
                                        }
                                        nodes {
                                            slug
                                        }
                                    }
                                }
                            }
                        }
                    ''',
                    'variables': {
                        'organization': self._organization,
                        'cursor': cursor,
                        'first': self._page_size,
                    },
                },
                headers={
                    'Authorization': f'Bearer {access_token}',
                },
            )
            response.raise_for_status()

            viewer = response.json()['data']['viewer']
            teams = viewer['organization']['teams']
            nodes = teams['nodes']
            edges = teams['edges']

            yield viewer['login']

            for team in nodes:
                yield team['slug']

            if len(nodes) < self._page_size:
                break

            cursor = edges[-1]['cursor']


class Auth:
    def insert(self, client_id: str, domain: str, owner: dict) -> None:
        raise NotImplementedError  # pragma: no cover

    def delete(self, client_id: str, domain: str) -> bool:
        raise NotImplementedError  # pragma: no cover

    def is_owner(self, domain: str, user: dict) -> bool:
        raise NotImplementedError  # pragma: no cover

    def client_id_for(self, domain: str) -> Optional[str]:
        raise NotImplementedError  # pragma: no cover

    def domain_for(self, client_id: str) -> Optional[str]:
        raise NotImplementedError  # pragma: no cover

    def domains(self) -> Iterable[str]:
        raise NotImplementedError  # pragma: no cover


class AzureAuth(Auth, LogMixin):
    def __init__(self, storage: AzureObjectStorage, sudo_scope: str) -> None:
        self._storage = storage
        self._sudo_scope = sudo_scope

    def insert(self, client_id: str, domain: str, owner: dict) -> None:
        auth = {'client_id': client_id, 'owner': owner['name'], 'domain': domain}
        self._storage.store_object(self._client_id_file(client_id), auth)
        self._storage.store_object(self._domain_file(domain), auth)
        self.log_info('Registered client %s at domain %s', client_id, domain)

    def is_owner(self, domain: str, user: dict) -> bool:
        if self._sudo_scope in user.get('scopes', []):
            return True

        try:
            auth = self._storage.fetch_object(self._domain_file(domain))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized domain %s', domain)
            return False

        return auth.get('owner') == user['name']

    def delete(self, client_id: str, domain: str) -> bool:
        self._storage.delete(self._domain_file(domain))
        self._storage.delete(self._client_id_file(client_id))
        return True

    def client_id_for(self, domain: str) -> Optional[str]:
        try:
            auth = self._storage.fetch_object(self._domain_file(domain))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized domain %s', domain)
            return None

        client_id = auth['client_id']

        self.log_debug('Domain %s has client %s', domain, client_id)
        return client_id

    def domain_for(self, client_id: str) -> Optional[str]:
        try:
            auth = self._storage.fetch_object(self._client_id_file(client_id))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized client %s', client_id)
            return None
        else:
            domain = auth['domain']
            self.log_debug('Client %s has domain %s', client_id, domain)
            return domain

    def domains(self) -> Iterable[str]:
        return self._storage.iter(self._domain_file(''))

    @classmethod
    def _domain_file(cls, domain: str) -> str:
        return f'domain/{domain}'

    @classmethod
    def _client_id_file(cls, client_id: str) -> str:
        return f'client_id/{client_id}'


class NoAuth(Auth):
    def __init__(self, client_id: str = 'service', domain: str = 'service'):
        self._client_id = client_id
        self._domain = domain

    def insert(self, client_id: str, domain: str, owner: dict) -> None:
        pass

    def delete(self, client_id: str, domain: str) -> bool:
        return False

    def client_id_for(self, domain: str) -> Optional[str]:
        return self._client_id

    def domains(self) -> Iterable[str]:
        return []

    def is_owner(self, domain: str, user: dict) -> bool:
        return True

    def domain_for(self, client_id: str) -> str:
        return self._domain



from abc import ABCMeta
from abc import abstractmethod
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Set
from typing import Union
from uuid import uuid4


class EmailStore(metaclass=ABCMeta):
    def __init__(self, restricted: Optional[Dict[str, Set[str]]] = None):
        self._restricted = restricted or {}

    def create(self, emails_or_attachments: Iterable[dict]):
        self._create((_add_uid(email_or_attachment) for email_or_attachment in emails_or_attachments
                      if not _is_restricted(email_or_attachment, self._restricted)))

    @abstractmethod
    def _create(self, emails_or_attachments: Iterable[dict]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get(self, uid: str) -> Optional[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get_attachment(self, email_id: str, attachment_id: str) -> Optional[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def inbox(self, email_address: str, page: int) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def outbox(self, email_address: str, page: int) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def sent(self, email_address: str, page: int) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def search(self, email_address: str, page: int, query: Optional[str]) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def pending(self, page: Optional[int]) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def has_unread(self, email_address: str) -> bool:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def num_pending(self) -> int:
        raise NotImplementedError  # pragma: no cover

    def mark_sent(self, emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._mark_sent(uids)

    def mark_read(self, email_address: str, emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._mark_read(email_address, uids)

    def delete(self, email_address: str, emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._delete(email_address, uids)

    @abstractmethod
    def _delete(self, email_address: str, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _mark_sent(self, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _mark_read(self, email_address: str, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover


def _is_restricted(email: dict, restricted: Dict[str, Set[str]]) -> bool:
    type_ = email.get('_type')
    if type_ and type_ != 'email':
        return False

    sender = email.get('from', '')
    recipients = _get_recipients(email)

    return any(restricted_inbox in recipients and sender not in allowed_senders
               for restricted_inbox, allowed_senders in restricted.items())


def _get_uid(email_or_uid: Union[str, dict]) -> str:
    try:
        return email_or_uid['_uid']
    except TypeError:
        return email_or_uid


def _get_recipients(email: dict) -> Set[str]:
    recipients = set()
    recipients.update(email.get('to', []))
    recipients.update(email.get('cc', []))
    recipients.update(email.get('bcc', []))
    return recipients


def _add_uid(email: dict) -> dict:
    email.setdefault('_uid', str(uuid4()))
    for attachment in email.get('attachments', []):
        attachment.setdefault('_uid', str(uuid4()))
    return email



from itertools import count
from mimetypes import guess_type
from time import sleep
from typing import Callable

from cached_property import cached_property
from python_http_client import BadRequestsError
from requests import delete as http_delete
from requests import get as http_get
from requests import post as http_post
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Attachment
from sendgrid.helpers.mail import Content
from sendgrid.helpers.mail import Email
from sendgrid.helpers.mail import Mail
from sendgrid.helpers.mail import MailSettings
from sendgrid.helpers.mail import Personalization
from sendgrid.helpers.mail import SandBoxMode

from opwen_email_server.constants.sendgrid import INBOX_URL
from opwen_email_server.constants.sendgrid import MAILBOX_CREATE_URL
from opwen_email_server.constants.sendgrid import MAILBOX_DETAIL_URL
from opwen_email_server.utils.log import LogMixin
from opwen_email_server.utils.serialization import to_base64


class SendSendgridEmail(LogMixin):
    def __init__(self, key: str, sandbox: bool = False) -> None:
        self._key = key
        self._sandbox = sandbox

    @cached_property
    def _client(self) -> Callable[[Mail], int]:
        if not self._key:

            def send_email_fake(email: Mail) -> int:
                self.log_warning('No key, not sending email %r', email)
                return 202

            return send_email_fake

        client = SendGridAPIClient(api_key=self._key)

        def send_email(email: Mail) -> int:
            if self._sandbox:
                self.log_warning('Sandbox mode, not delivering email %s', email)
                email.mail_settings = MailSettings()
                email.mail_settings.sandbox_mode = SandBoxMode()
                email.mail_settings.sandbox_mode.enable = True

            response = client.send(email.get())
            return response.status_code

        return send_email

    def __call__(self, email: dict) -> bool:
        email_id = email.get('_uid', '')
        email = self._create_email(email, email_id)
        return self._send_email(email, email_id)

    def _send_email(self, email: Mail, email_id: str) -> bool:
        self.log_debug('about to send email %s', email_id)
        try:
            status = self._client(email)
        except BadRequestsError as ex:
            status = ex.status_code
            errors = ex.to_dict.get('errors', [])
            self.log_exception(ex, 'error sending email %s:%s:%r', email_id, email, errors)
        except Exception as ex:
            status = getattr(ex, 'code', -1)
            self.log_exception(ex, 'error sending email %s:%s', email_id, email)
        else:
            self.log_info('sent email %s', email_id)

        return status in (200, 201, 202)

    def _create_email(self, email: dict, email_id: str) -> Mail:
        self.log_debug('converting email %s to sendgrid format', email_id)
        mail = Mail()
        personalization = Personalization()

        for i, to in enumerate(email.get('to', [])):
            personalization.add_to(Email(to))
            self.log_debug('added to %d to email %s', i, email_id)

        for i, cc in enumerate(email.get('cc', [])):
            personalization.add_cc(Email(cc))
            self.log_debug('added cc %d to email %s', i, email_id)

        for i, bcc in enumerate(email.get('bcc', [])):
            personalization.add_bcc(Email(bcc))
            self.log_debug('added bcc %d to email %s', i, email_id)

        mail.add_personalization(personalization)
        self.log_debug('added recipients to email %s', email_id)

        mail.subject = email.get('subject', '(no subject)')
        self.log_debug('added subject to email %s', email_id)

        mail.add_content(Content('text/html', email.get('body', '(no content)')))
        self.log_debug('added content to email %s', email_id)

        # at some point SendGrid had the ability to send from subdomains of a verified
        # domain, so verifying {domain} let us send from {client}.{domain}
        # ...this feature went away so the following is a work-around which
        # changes the from address to the verified root domain but using reply-to
        # so that the email still gets routed back to the original client
        # ...this is a pretty ugly hack and the real fix is to change the logic of
        # how we sign up users on clients to use the new format {user}-{client}@{domain}
        from_email = email.get('from')
        if from_email:
            user, client_domain = from_email.split('@')
            client_domain_parts = client_domain.split('.')
            if len(client_domain_parts) == 3:
                client = client_domain_parts[0]
                domain = '.'.join(client_domain_parts[1:])
                mail.from_email = Email('{}-{}@{}'.format(user, client, domain))
                mail.reply_to = Email(from_email)
            else:
                mail.from_email = Email(from_email)
            self.log_debug('added from to email %s', email_id)

        for i, attachment in enumerate(email.get('attachments', [])):
            mail.add_attachment(self._create_attachment(attachment))
            self.log_debug('added attachment %d to email %s', i, email_id)

        self.log_debug('converted email %s to sendgrid format', email_id)
        return mail

    @classmethod
    def _create_attachment(cls, attachment: dict) -> Attachment:
        filename = attachment.get('filename', '')
        content = attachment.get('content', b'')

        return Attachment(
            disposition='attachment',
            file_name=filename,
            content_id=filename,
            file_type=guess_type(filename)[0],
            file_content=to_base64(content),
        )


class _SendgridManagement(LogMixin):
    def __init__(self, key: str) -> None:
        self._key = key

    def __call__(self, client_id: str, domain: str) -> None:
        if not self._key:
            self.log_warning('No key, skipping mailbox setup for %s', domain)
            return

        self._run(client_id, domain)

    def _run(self, client_id: str, domain: str) -> None:
        raise NotImplementedError  # pragma: no cover


class DeleteSendgridMailbox(_SendgridManagement):
    def _run(self, client_id: str, domain: str) -> None:
        response = http_delete(
            url=MAILBOX_DETAIL_URL.format(domain),
            headers={
                'Authorization': f'Bearer {self._key}',
            },
        )

        if not response.ok and response.status_code != 404:
            response.raise_for_status()

        self.log_info('Deleted mailbox for %s', domain)


class SetupSendgridMailbox(_SendgridManagement):
    def __init__(self, key: str, max_retries: int, retry_interval_seconds: float):
        super().__init__(key)
        self._max_retries = max_retries
        self._retry_interval_seconds = retry_interval_seconds

    def _run(self, client_id: str, domain: str) -> None:
        for retry in count():
            get_response = http_get(
                url=MAILBOX_DETAIL_URL.format(domain),
                headers={
                    'Authorization': f'Bearer {self._key}',
                },
            )

            if get_response.ok:
                self.log_debug('Mailbox %s already exists', domain)
                break

            create_response = http_post(
                url=MAILBOX_CREATE_URL,
                json={
                    'hostname': domain,
                    'url': INBOX_URL.format(client_id),
                    'spam_check': True,
                    'send_raw': True,
                },
                headers={
                    'Authorization': f'Bearer {self._key}',
                },
            )

            if create_response.ok:
                self.log_info('Set up mailbox for %s', domain)
                break

            if retry > self._max_retries:
                self.log_warning('Too many attempts to set up mailbox for %s', domain)
                create_response.raise_for_status()

            sleep(self._retry_interval_seconds)



from functools import lru_cache
from itertools import islice
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import TypeVar

T = TypeVar('T')


def to_iterable(obj: Optional[T]) -> Iterable[T]:
    if obj:
        yield obj


def chunks(iterable: Iterable[T], chunk_size: int) -> Iterable[Iterable[T]]:
    it = iter(iterable)
    while True:
        chunk = tuple(islice(it, chunk_size))
        if not chunk:
            return
        yield chunk


def singleton(func: Callable) -> Callable:
    return lru_cache(maxsize=1)(func)


def append(iterable: Iterable[T], next_item: T) -> Iterable[T]:
    for item in iterable:
        yield item
    yield next_item



from opwen_email_server.mailers.echo import ECHO_ADDRESS
from opwen_email_server.mailers.echo import EchoEmailFormatter
from opwen_email_server.mailers.wikipedia import WIKIPEDIA_ADDRESS
from opwen_email_server.mailers.wikipedia import WikipediaEmailFormatter

REGISTRY = {
    ECHO_ADDRESS: EchoEmailFormatter(),
    WIKIPEDIA_ADDRESS: WikipediaEmailFormatter(),
}



from flask_security import LoginForm as _LoginForm
from flask_security import RegisterForm as _RegisterForm
from flask_security.forms import email_required
from flask_security.forms import email_validator
from flask_security.forms import unique_user_email
from wtforms import IntegerField
from wtforms.validators import NoneOf
from wtforms.validators import Regexp

from opwen_email_client.util.wtforms import SuffixedStringField
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n


# noinspection PyClassHasNoInit
class LoginForm(_LoginForm):
    email = SuffixedStringField(suffix='@{}'.format(AppConfig.CLIENT_EMAIL_HOST))


email_character_validator = Regexp('^[a-zA-Z0-9-.@]*$', message=i8n.EMAIL_CHARACTERS)

forbidden_account_validator = NoneOf(AppConfig.FORBIDDEN_ACCOUNTS, message=i8n.FORBIDDEN_ACCOUNT)


# noinspection PyClassHasNoInit
class RegisterForm(_RegisterForm):
    email = SuffixedStringField(suffix='@{}'.format(AppConfig.CLIENT_EMAIL_HOST),
                                validators=[
                                    email_character_validator,
                                    forbidden_account_validator,
                                    email_required,
                                    email_validator,
                                    unique_user_email,
                                ])

    timezone_offset_minutes = IntegerField(default=0)



from datetime import datetime
from typing import Callable

from opwen_email_server.utils.log import LogMixin

ECHO_ADDRESS = 'echo@bot.lokole.ca'


class EchoEmailFormatter(LogMixin):
    def __init__(self, now: Callable[[], datetime] = datetime.utcnow):
        self._now = now

    def __call__(self, email: dict) -> dict:
        email['to'] = [email['from']]
        email['from'] = ECHO_ADDRESS
        email['sent_at'] = self._now().strftime('%Y-%m-%d %H:%M')
        return email



from urllib.parse import unquote
from urllib.parse import urlparse

import click
from azure.servicebus import ServiceBusClient
from libcloud.storage.providers import get_driver

from opwen_email_server import config

_STORAGES = (
    (
        config.BLOBS_ACCOUNT,
        config.BLOBS_KEY,
        config.BLOBS_HOST,
        config.BLOBS_SECURE,
    ),
    (
        config.TABLES_ACCOUNT,
        config.TABLES_KEY,
        config.TABLES_HOST,
        config.TABLES_SECURE,
    ),
    (
        config.CLIENT_STORAGE_ACCOUNT,
        config.CLIENT_STORAGE_KEY,
        config.CLIENT_STORAGE_HOST,
        config.CLIENT_STORAGE_SECURE,
    ),
)


@click.group()
def cli():
    pass


@cli.command()
@click.option('--separator', '-s', default='\n')
def print_queues(separator):
    click.echo(
        separator.join((
            config.REGISTER_CLIENT_QUEUE,
            config.INBOUND_STORE_QUEUE,
            config.WRITTEN_STORE_QUEUE,
            config.PROCESS_SERVICE_QUEUE,
            config.SEND_QUEUE,
            config.MAILBOX_RECEIVED_QUEUE,
            config.MAILBOX_SENT_QUEUE,
        )))


@cli.command()
@click.option('-s', '--suffix', default='')
def delete_queues(suffix):
    suffix = suffix.replace('-', '_')

    queue_broker = urlparse(config.QUEUE_BROKER)
    if queue_broker.scheme != 'azureservicebus':
        click.echo(f'Skipping queue cleanup for {queue_broker.scheme}')
        return

    client = ServiceBusClient(
        service_namespace=queue_broker.hostname,
        shared_access_key_name=unquote(queue_broker.username),
        shared_access_key_value=unquote(queue_broker.password),
    )

    for queue in client.list_queues():
        if queue.name.endswith(suffix):
            click.echo(f'Deleting queue {queue.name}')
            client.delete_queue(queue.name)


@cli.command()
@click.option('-s', '--suffix')
def delete_containers(suffix):
    for account, key, host, secure in _STORAGES:
        client = get_driver(config.STORAGE_PROVIDER)(account, key, host=host, secure=secure)

        for container in client.iterate_containers():
            if container.name.endswith(suffix):
                click.echo(f'Deleting container {container.name}')

                response = client.connection.request(
                    f'/{container.name}',
                    params={'restype': 'container'},
                    method='DELETE',
                )

                if response.status != 202:
                    raise ValueError(f'Unable to delete container {container.name}')


if __name__ == '__main__':
    cli()



from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from cached_property import cached_property
from flask_security import LoginForm
from libcloud.storage.types import ObjectDoesNotExistError

from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.domain.email.sync import Sync
from opwen_email_client.domain.email.user_store import User
from opwen_email_client.domain.email.user_store import UserReadStore
from opwen_email_client.domain.email.user_store import UserStore
from opwen_email_client.domain.email.user_store import UserWriteStore
from opwen_email_client.webapp.config import AppConfig
from opwen_email_server.constants import mailbox
from opwen_email_server.integration.azure import get_email_storage
from opwen_email_server.integration.azure import get_mailbox_storage
from opwen_email_server.integration.azure import get_pending_storage
from opwen_email_server.integration.azure import get_user_storage
from opwen_email_server.integration.celery import send_and_index_email
from opwen_email_server.services.storage import AzureObjectStorage
from opwen_email_server.services.storage import AzureTextStorage
from opwen_email_server.utils.collections import chunks
from opwen_email_server.utils.email_parser import descending_timestamp
from opwen_email_server.utils.email_parser import ensure_has_sent_at
from opwen_email_server.utils.email_parser import get_domain
from opwen_email_server.utils.email_parser import get_recipients
from opwen_email_server.utils.log import LogMixin


class AzureRole:
    def __init__(self):
        raise NotImplementedError


class AzureUser(User):
    def __init__(self, **data):
        super().__setattr__('_data', data)

    def __getattr__(self, item):
        data = super().__getattribute__('_data')
        return data.get(item)

    def __setattr__(self, key, value):
        data = super().__getattribute__('_data')
        data[key] = value

    @property
    def id(self) -> Union[str, int]:
        return self.email

    @property
    def email(self) -> str:
        return super().__getattribute__('_data')['email']

    @property
    def password(self) -> str:
        return super().__getattribute__('_data')['password']

    @property
    def roles(self) -> List[str]:
        return []

    @property
    def active(self) -> bool:
        return True

    @property
    def is_admin(self) -> bool:
        return False


class AzureUserStore(UserStore, UserReadStore, UserWriteStore):
    def __init__(self, user_storage: AzureObjectStorage):
        UserReadStore.__init__(self, user_model=AzureUser, role_model=AzureRole)
        UserWriteStore.__init__(self, db=None)
        UserStore.__init__(self, read=self, write=self)
        self._user_storage = user_storage

    def init_app(self, app):
        pass

    def fetch_all(self, user: AzureUser) -> List[AzureUser]:
        domain_users = []
        for email in self._user_storage.iter(f'{get_domain(user.email)}/'):
            domain_user = self.get_user(email)
            if domain_user is not None:
                domain_users.append(domain_user)
        return domain_users

    def fetch_pending(self) -> List[AzureUser]:
        return []

    def get_user(self, id_or_email) -> Optional[AzureUser]:
        try:
            data = self._user_storage.fetch_object(self._path_for(id_or_email))
        except ObjectDoesNotExistError:
            return None
        else:
            return AzureUser(**data)

    def find_user(self, *args, **kwargs) -> Optional[AzureUser]:
        if 'id' not in kwargs and 'email' not in kwargs:
            raise NotImplementedError(f'Unable to find_user by: {", ".join(kwargs.keys())}')

        id_or_email = kwargs.get('email') or kwargs.get('id')
        return self.get_user(id_or_email)

    def find_role(self, *args, **kwargs):
        raise NotImplementedError

    def put(self, user: AzureUser) -> AzureUser:
        return user

    def commit(self) -> None:
        pass

    def delete(self, user: AzureUser) -> None:
        self._user_storage.delete(self._path_for(user.email))

    @classmethod
    def _path_for(cls, email: str) -> str:
        return f'{get_domain(email)}/{email}'


class AzureEmailStore(EmailStore, LogMixin):
    def __init__(self, email_storage: AzureObjectStorage, mailbox_storage: AzureTextStorage,
                 pending_storage: AzureTextStorage, send_email: Callable[[str], None]):
        super().__init__(restricted=None)
        self._email_storage = email_storage
        self._mailbox_storage = mailbox_storage
        self._pending_storage = pending_storage
        self._send_email = send_email

    def _create(self, emails_or_attachments: Iterable[dict]):
        for email in emails_or_attachments:
            if email.get('_type') == 'attachment':
                raise NotImplementedError

            domain = get_domain(email.get('from', ''))
            if not domain.endswith(mailbox.MAILBOX_DOMAIN):
                continue

            ensure_has_sent_at(email)
            email.pop('csrf_token', None)
            email_id = email['_uid']
            self._email_storage.store_object(email_id, email)
            self._pending_storage.store_text(f'{domain}/{email_id}', 'pending')
            self._send_email(email_id)

    def get(self, uid: str) -> Optional[dict]:
        try:
            email = self._email_storage.fetch_object(uid)
        except ObjectDoesNotExistError:
            self.log_warning('Email at %s does not exist', uid)
            return None
        else:
            email['read'] = True
            for i, attachment in enumerate(email.get('attachments', [])):
                attachment.setdefault('_uid', i)
                attachment.setdefault('cid', None)
            return email

    def get_attachment(self, email_id: str, attachment_id: str) -> Optional[dict]:
        email = self.get(email_id)
        if not email:
            return None

        attachments = email.get('attachments', [])
        for attachment in attachments:
            if attachment.get('_uid') == attachment_id:
                return attachment

        try:
            return attachments[int(attachment_id)]
        except (IndexError, ValueError):
            return None

    def inbox(self, email_address: str, page: int) -> Iterable[dict]:
        return self._iter_mailbox(email_address, page, mailbox.RECEIVED_FOLDER)

    def sent(self, email_address: str, page: int) -> Iterable[dict]:
        return self._iter_mailbox(email_address, page, mailbox.SENT_FOLDER)

    def _iter_mailbox(self, email_address: str, page: int, folder: str) -> Iterable[dict]:
        domain = get_domain(email_address)
        emails = self._mailbox_storage.iter(f'{domain}/{email_address}/{folder}')
        for i, resource_ids in enumerate(chunks(emails, AppConfig.EMAILS_PER_PAGE), start=1):
            if i != page:
                continue
            for resource_id in resource_ids:
                email_id = resource_id.split('/')[-1]
                email = self.get(email_id)
                if email:
                    yield email

    def search(self, email_address: str, page: int, query: Optional[str]) -> Iterable[dict]:
        return []

    def outbox(self, email_address: str, page: int) -> Iterable[dict]:
        return []

    def pending(self, page: Optional[int]) -> Iterable[dict]:
        return []

    def has_unread(self, email_address: str) -> bool:
        return False

    def num_pending(self) -> int:
        return 0

    def _delete(self, email_address: str, uids: Iterable[str]):
        domain = get_domain(email_address)

        for uid in uids:
            email = self.get(uid)
            if not email:
                continue

            if email_address == email.get('from'):
                folder = mailbox.SENT_FOLDER
            elif email_address in get_recipients(email):
                folder = mailbox.RECEIVED_FOLDER
            else:
                continue

            desc_prefix = descending_timestamp(email['sent_at'])

            self._mailbox_storage.delete(f"{domain}/{email_address}/{folder}/{desc_prefix}/{uid}")

    def _mark_sent(self, uids: Iterable[str]):
        pass

    def _mark_read(self, email_address: str, uids: Iterable[str]):
        pass


class NoSync(Sync):
    def upload(self, items: Iterable, users: Iterable[User]) -> Iterable[str]:
        return []

    def download(self) -> Iterable:
        return []


class AzureIoc:
    @cached_property
    def email_store(self):
        return AzureEmailStore(
            email_storage=get_email_storage(),
            mailbox_storage=get_mailbox_storage(),
            pending_storage=get_pending_storage(),
            send_email=send_and_index_email,
        )

    @cached_property
    def email_sync(self):
        return NoSync()

    @cached_property
    def user_store(self):
        return AzureUserStore(user_storage=get_user_storage())

    @cached_property
    def login_form(self):
        return LoginForm



from datetime import timedelta
from os.path import splitext

from bs4 import BeautifulSoup
from flask import url_for

from opwen_email_client.webapp import app


@app.template_filter('asset_url')
def asset_url(asset_path: str) -> str:
    if app.config['TESTING']:
        return url_for('static', filename=asset_path)

    asset_path, extension = splitext(asset_path)
    return url_for('static', filename='{}.min{}'.format(asset_path, extension))


@app.template_filter('render_body')
def render_body(email: dict) -> str:
    body = email.get('body')
    if not body:
        return ''

    body = body.replace('\n', '<br>')

    soup = BeautifulSoup(body, 'html.parser')
    images = soup.find_all('img')
    if not images:
        return body

    attachments = {attachment['cid']: attachment['_uid'] for attachment in email.get('attachments', [])}
    for img in images:
        src = img.get('src')
        if not src:
            continue
        if src.startswith('cid:'):
            attachment_cid = src[4:]
            attachment_id = attachments.get(attachment_cid)
            if attachment_id:
                src = url_for('download_attachment', email_id=email['_uid'], attachment_id=attachment_id)
        del img['src']
        img['data-original'] = src
    body = str(soup)

    return body


@app.context_processor
def _inject_format_last_login():
    def format_last_login(user, current_user) -> str:
        if not user.last_login_at:
            return ''

        date = user.last_login_at - timedelta(minutes=current_user.timezone_offset_minutes)
        return date.strftime('%Y-%m-%d %H:%M')

    return {'format_last_login': format_last_login}



from abc import ABCMeta
from abc import abstractmethod
from base64 import b64decode
from base64 import b64encode
from copy import deepcopy
from json import dumps
from json import loads
from typing import TypeVar

T = TypeVar('T')


class Serializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, obj: T, type_: str = '') -> bytes:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def deserialize(self, serialized: bytes, type_: str = '') -> T:
        raise NotImplementedError  # pragma: no cover


class JsonSerializer(Serializer):
    _encoding = 'utf-8'
    _separators = (',', ':')

    def serialize(self, obj: dict, type_: str = '') -> bytes:
        if not type_ or type_ == 'email':
            obj = self._encode_attachments(obj)
        elif type_ == 'attachment':
            self._encode_attachment(obj)

        serialized = dumps(obj, separators=self._separators, sort_keys=True)
        return serialized.encode(self._encoding)

    def deserialize(self, serialized: bytes, type_: str = '') -> dict:
        decoded = serialized.decode(self._encoding)
        obj = loads(decoded)

        if not type_ or type_ == 'email':
            email = obj
            email = self._decode_attachments(email)
            return email
        elif type_ == 'attachment':
            attachment = obj
            self._decode_attachment(attachment)
            return attachment

    @classmethod
    def _encode_attachments(cls, email: dict) -> dict:
        attachments = email.get('attachments', [])
        if not attachments:
            return email

        email = deepcopy(email)
        for attachment in email['attachments']:
            cls._encode_attachment(attachment)
        return email

    @classmethod
    def _encode_attachment(cls, attachment: dict) -> None:
        content = attachment.get('content', b'')
        if content:
            attachment['content'] = b64encode(content).decode('ascii')

    @classmethod
    def _decode_attachments(cls, email: dict) -> dict:
        attachments = email.get('attachments', [])
        if not attachments:
            return email

        email = deepcopy(email)
        for attachment in email['attachments']:
            cls._decode_attachment(attachment)
        return email

    @classmethod
    def _decode_attachment(cls, attachment: dict) -> None:
        content = attachment.get('content', '')
        if content:
            attachment['content'] = b64decode(content)



from typing_extensions import Final  # noqa: F401

MAILBOX_DOMAIN = 'lokole.ca'  # type: Final
RECEIVED_FOLDER = 'received'  # type: Final
SENT_FOLDER = 'sent'  # type: Final
FUTURE_TIMESTAMP = 2100000000  # type: Final



from datetime import datetime
from typing import Callable
from urllib.parse import urlparse

from requests import get
from wikipedia import WikipediaPage
from wikipedia import languages
from wikipedia import page
from wikipedia import set_lang
from wikipedia.exceptions import DisambiguationError
from wikipedia.exceptions import PageError

from opwen_email_server.utils.log import LogMixin

WIKIPEDIA_ADDRESS = 'wikipedia@bot.lokole.ca'


class WikipediaEmailFormatter(LogMixin):
    def __init__(self,
                 languages_getter: Callable[[], dict] = languages,
                 language_setter: Callable[[str], None] = set_lang,
                 page_fetch: Callable[[str], WikipediaPage] = page,
                 now: Callable[[], datetime] = datetime.utcnow):
        self._now = now
        self._languages = languages_getter
        self._language_setter = language_setter
        self._page_fetch = page_fetch

    def _get_download_link(self, url: str) -> str:
        parsed_url = urlparse(url)
        return parsed_url.scheme + '://' + parsed_url.netloc + '/api/rest_v1/page/pdf/' + parsed_url.path.split('/')[-1]

    def __call__(self, email: dict) -> dict:
        language = email['subject']
        if language in self._languages():
            self._language_setter(language)
        else:
            self._language_setter('en')

        try:
            wiki_page = self._page_fetch(email['body'].strip())
        except DisambiguationError as e:
            subject = 'Suggested Searches'
            body = 'Multiple results found. Try again with the following: \n{}'.format('\n'.join(e.options))
        except PageError:
            subject = 'No results'
            body = 'No results found for: {}'.format(email['body'])
        else:
            subject = wiki_page.title
            body = 'Results found'
            download_link = self._get_download_link(wiki_page.url)
            pdf_file = get(download_link)
            email['attachments'] = [{
                'filename': subject + '.pdf',
                'content': pdf_file.content,
            }]

        email['to'] = [email['from']]
        email['from'] = WIKIPEDIA_ADDRESS
        email['sent_at'] = self._now().strftime('%Y-%m-%d %H:%M')
        email['subject'] = subject
        email['body'] = body

        return email



from base64 import b64decode
from base64 import b64encode
from gzip import GzipFile
from io import BytesIO
from json import JSONDecodeError
from json import dumps
from json import loads
from typing import Optional

from msgpack import packb as msgpack_dump
from msgpack import unpackb as msgpack_load


def to_json(obj: object) -> str:
    return dumps(obj, separators=(',', ':'))


def from_json(obj: str) -> dict:
    return loads(obj)


def to_jsonl_bytes(obj) -> bytes:
    return to_json(obj).encode('utf-8') + b'\n'


def from_jsonl_bytes(obj: bytes) -> Optional[dict]:
    try:
        decoded = obj.decode('utf-8')
    except UnicodeDecodeError:
        return None
    decoded = decoded.rstrip(',\n')
    try:
        return from_json(decoded)
    except JSONDecodeError:
        return None


def to_msgpack_bytes(obj) -> bytes:
    encoded = msgpack_dump(obj, use_bin_type=True)
    return encoded + b'\n'


def from_msgpack_bytes(serialized: bytes) -> dict:
    encoded = serialized.rstrip(b'\n')
    return msgpack_load(encoded, raw=False)


def to_base64(content: bytes) -> str:
    return b64encode(content).decode('ascii')


def from_base64(encoded: str) -> bytes:
    return b64decode(encoded)


def gzip_bytes(uncompressed_bytes: bytes) -> bytes:
    stream = BytesIO()
    with GzipFile(fileobj=stream, mode='wb') as fobj:
        fobj.write(uncompressed_bytes)  # type: ignore
    stream.seek(0)
    compressed = stream.read()
    return compressed


def gunzip_bytes(compressed: bytes) -> bytes:
    stream = BytesIO()
    stream.write(compressed)
    stream.seek(0)
    with GzipFile(fileobj=stream, mode='rb') as fobj:
        uncompressed_bytes = fobj.read()  # type: ignore
    return uncompressed_bytes



from contextlib import contextmanager
from importlib import import_module
from json import loads
from logging import Logger
from pathlib import Path
from subprocess import check_call  # nosec
from subprocess import check_output  # nosec
from sys import executable
from time import sleep
from typing import Mapping
from typing import Optional

from cached_property import cached_property
from flask import render_template
from requests import get
from requests.exceptions import HTTPError

from opwen_email_client.domain import sim
from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.domain.email.sync import Sync
from opwen_email_client.domain.email.user_store import UserStore
from opwen_email_client.domain.modem import e303
from opwen_email_client.domain.modem import e353
from opwen_email_client.domain.modem import e3131
from opwen_email_client.domain.modem import modem_is_plugged
from opwen_email_client.domain.modem import modem_is_setup
from opwen_email_client.domain.modem import setup_modem
from opwen_email_client.domain.sim import dialup
from opwen_email_client.util.os import backup
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n
from opwen_email_client.webapp.config import root_domain


class SyncEmails(object):
    def __init__(self, email_store: EmailStore, email_sync: Sync, user_store: UserStore, log: Logger):
        self._email_store = email_store
        self._email_sync = email_sync
        self._user_store = user_store
        self._log = log

    def _upload(self):
        pending = self._email_store.pending(page=None)
        users = self._user_store.fetch_pending()

        # noinspection PyBroadException
        try:
            uploaded = self._email_sync.upload(pending, users)
        except Exception:
            self._log.exception('Unable to upload emails')
        else:
            if uploaded:
                self._email_store.mark_sent(uploaded)
                self._mark_as_synced(users)

    def _mark_as_synced(self, users):
        for user in users:
            user.synced = True
            self._user_store.w.put(user)
        self._user_store.w.commit()

    def _download(self):
        # noinspection PyBroadException
        try:
            downloaded = self._email_sync.download()
        except Exception:
            self._log.exception('Unable to download emails')
        else:
            self._email_store.create(downloaded)

    def _sync(self):
        self._upload()
        self._download()

    def __call__(self):
        self._sync()


class UpdateCode(object):
    _package_name = 'opwen_email_client'

    def __init__(self, version: Optional[str], log: Logger):
        self._version = version
        self._log = log

    def __call__(self):
        if self._version:
            package = '{}=={}'.format(self._package_name, self._version)
            self._log.debug('Updating to version %s', self._version)
        else:
            package = self._package_name
            self._log.debug('Updating to latest version')

        stdout = check_output([executable, '-m', 'pip', 'install', '-U', package])

        self._log.debug('Pip install log: %s', stdout)


class RestartApp(object):
    def __init__(self, restart_paths: Mapping[str, str]):
        self._restart_paths = restart_paths

    def __call__(self):
        for path, signal in self._restart_paths.items():
            path = Path(path)
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text(signal, encoding='ascii')


class RestartAppComponent(object):
    def __init__(self, restart_path: str):
        self._restart_path = restart_path

    def __call__(self):
        path = Path(self._restart_path)

        if not path.is_file():
            return

        component_name = path.name
        signal = path.read_text(encoding='ascii').strip()

        if signal:
            check_call(['supervisorctl', 'signal', signal, component_name])
        else:
            check_call(['supervisorctl', 'restart', component_name])

        path.unlink()


class SendWelcomeEmail(object):
    def __init__(self, to: str, time, email_store: EmailStore):
        self._to = to
        self._time = time
        self._email_store = email_store

    def __call__(self, *args, **kwargs):
        email_body = render_template('emails/account_finalized.html', email=self._to)
        self._email_store.create([{
            'sent_at': self._time.strftime("%Y-%m-%d %H:%M"),
            'to': [self._to],
            'from': 'info@team.lokole.ca',
            'subject': i8n.WELCOME,
            'body': email_body,
        }])


class StartInternetConnection(object):
    _supported_modems = (e303, e353, e3131)

    def __init__(self, modem_config_dir: str, sim_config_dir: str, sim_type: str, state_dir: str):
        self._modem_config_dir = Path(modem_config_dir)
        self._sim_config_dir = Path(sim_config_dir)
        self._sim_type = sim_type
        self._state_dir = Path(state_dir)
        self._modem_target_mode = '1506'

    @cached_property
    def _wvdial_config(self) -> Path:
        wvdial_config = self._sim_config_dir / self._sim_type
        if wvdial_config.is_file():
            return wvdial_config

        sim_config_module = '{}.{}'.format(sim.__name__, self._sim_type)
        try:
            sim_config = import_module(sim_config_module)
        except ImportError:
            raise Exception('SIM config {} does not exist'.format(wvdial_config))
        else:
            wvdial_config.parent.mkdir(parents=True, exist_ok=True)
            with wvdial_config.open('w', encoding='utf-8') as fobj:
                fobj.write(sim_config.wvdial)

        return wvdial_config

    @property
    def _wvdial_log(self) -> Path:
        wvdial_log = self._state_dir / 'wvdial.log'

        backup(wvdial_log)

        wvdial_log.write_bytes(b'')
        return wvdial_log

    def _modem_config_for(self, modem) -> Path:
        modem_config = self._modem_config_dir / modem.uid
        if modem_config.is_file():
            return modem_config

        modem_config.parent.mkdir(parents=True, exist_ok=True)
        with modem_config.open('w', encoding='utf-8') as fobj:
            fobj.write(modem.modeswitch)

        return modem_config

    def _setup_modem(self, poll_seconds: int):
        if not modem_is_plugged():
            raise Exception('No modem plugged in')

        if not modem_is_setup(self._modem_target_mode):
            for modem in self._supported_modems:
                if modem_is_plugged(modem):
                    self._modem_target_mode = modem.target
                    setup_modem(self._modem_config_for(modem))
                    break
            else:
                raise Exception('Unknown modem')

            while not modem_is_setup(self._modem_target_mode):
                sleep(poll_seconds)

    @contextmanager
    def __call__(self):
        connection = None

        if self._sim_type != 'Ethernet':
            self._setup_modem(poll_seconds=2)

            connection = dialup(
                self._wvdial_config,
                self._wvdial_log,
                max_retries=90,
                poll_seconds=1,
            )

        try:
            yield
        finally:
            if connection is not None:
                connection.terminate()


class ClientRegister(object):
    def __init__(self, client_name: str, access_token: str, path: str, logger: Logger):
        self._client_name = client_name
        self._github_access_token = access_token
        self._settings_path = path
        self._log = logger

    @property
    def client_domain(self):
        return '{}.{}'.format(self._client_name, 'lokole.ca')

    @property
    def server_endpoint(self):
        return AppConfig.EMAIL_SERVER_ENDPOINT or 'mailserver.lokole.ca'

    @property
    def client_url_details(self):
        return 'https://{}/api/email/register/{}'.format(self.server_endpoint, self.client_domain)

    def _format_settings(self):
        if AppConfig.RESTART_PATHS:
            restart_paths_list = ['{}={}'.format(key, value) for (key, value) in AppConfig.RESTART_PATHS.items()]
            restart_path = ','.join(restart_paths_list)
        else:
            restart_path = ''

        return {
            'OPWEN_APP_ROOT': AppConfig.APP_ROOT,
            'OPWEN_STATE_DIRECTORY': AppConfig.STATE_BASEDIR,
            'OPWEN_SESSION_KEY': AppConfig.SECRET_KEY,
            'OPWEN_MAX_UPLOAD_SIZE_MB': AppConfig.MAX_UPLOAD_SIZE_MB,
            'OPWEN_SIM_TYPE': AppConfig.SIM_TYPE,
            'OPWEN_EMAIL_SERVER_HOSTNAME': self.server_endpoint,
            'OPWEN_CLIENT_NAME': self._client_name,
            'OPWEN_ROOT_DOMAIN': root_domain,
            'OPWEN_RESTART_PATH': restart_path,
        }

    def _write_settings_to_file(self, client_values):
        client_values_list = ['{}={}'.format(key, value) for (key, value) in client_values.items()]

        with open(self._settings_path, 'w') as fobj:
            fobj.write('\n'.join(client_values_list))

    def __call__(self):
        while True:
            get_headers = {'Authorization': 'Bearer {}'.format(self._github_access_token)}

            get_response = get(self.client_url_details, headers=get_headers)
            if get_response.status_code == 404:
                continue

            try:
                get_response.raise_for_status()
            except HTTPError as ex:
                self._log.exception('Unable to fetch client {client_name}: [{status_code}] {message}'.format(
                    client_name=self._client_name,
                    status_code=get_response.status_code,
                    message=ex.read().decode('utf-8').strip()))
            else:
                client_info = loads(get_response.text)
                break

        opwen_settings = self._format_settings()
        registration_details = {
            'OPWEN_CLIENT_ID': client_info['client_id'],
            'OPWEN_REMOTE_ACCOUNT_NAME': client_info['storage_account'],
            'OPWEN_REMOTE_ACCOUNT_KEY': client_info['storage_key'],
            'OPWEN_REMOTE_RESOURCE_CONTAINER': client_info['resource_container'],
        }

        self._write_settings_to_file({**opwen_settings, **registration_details})



from hashlib import sha256
from random import Random
from typing import Optional
from uuid import UUID
from uuid import uuid4

from opwen_email_server.utils.serialization import to_msgpack_bytes


class NewGuid:
    def __init__(self, seed: Optional[int] = None) -> None:
        self._random = None  # type: Optional[Random]

        if seed is not None:
            self._random = Random()
            self._random.seed(seed)

    def __call__(self) -> str:
        if self._random is not None:
            guid = UUID(int=self._random.getrandbits(128))
        else:
            guid = uuid4()

        return str(guid)


def new_email_id(email: dict) -> str:
    return sha256(to_msgpack_bytes(email)).hexdigest()



from collections import namedtuple
from io import BytesIO
from tarfile import TarFile
from tempfile import NamedTemporaryFile
from typing import IO
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import Tuple

from cached_property import cached_property
from libcloud.storage.base import Container
from libcloud.storage.base import Object
from libcloud.storage.base import StorageDriver
from libcloud.storage.providers import get_driver
from libcloud.storage.types import ContainerAlreadyExistsError
from libcloud.storage.types import ContainerDoesNotExistError
from libcloud.storage.types import ObjectDoesNotExistError
from libcloud.storage.types import Provider
from xtarfile import open as tarfile_open
from xtarfile.xtarfile import SUPPORTED_FORMATS

from opwen_email_server.utils.log import LogMixin
from opwen_email_server.utils.serialization import from_msgpack_bytes
from opwen_email_server.utils.serialization import gunzip_bytes
from opwen_email_server.utils.serialization import gzip_bytes
from opwen_email_server.utils.serialization import to_msgpack_bytes
from opwen_email_server.utils.temporary import create_tempfilename
from opwen_email_server.utils.temporary import removing

AccessInfo = namedtuple('AccessInfo', ['account', 'key', 'container'])

Upload = Tuple[str, Iterable[dict], Callable[[dict], bytes]]
Download = Tuple[str, Callable[[bytes], dict]]


class _Container:
    def __init__(self, wrapped: Container):
        self._wrapped = wrapped

    def get_object(self, object_name: str) -> Object:
        return self._wrapped.get_object(object_name)

    def iterate_objects(self, prefix: Optional[str] = None) -> Iterable[Object]:
        return self._wrapped.iterate_objects(prefix)

    def upload_object(self, file_path: str, object_name: str) -> Object:
        return self._wrapped.upload_object(file_path, object_name)

    def upload_object_via_stream(self, iterator: Iterator[bytes], object_name: str) -> Object:
        return self._wrapped.upload_object_via_stream(iterator, object_name)


class _CaseInsensitiveContainer(_Container):
    def get_object(self, object_name: str) -> Object:
        object_name = object_name.lower()
        return super().get_object(object_name)

    def iterate_objects(self, prefix: Optional[str] = None) -> Iterable[Object]:
        prefix = prefix.lower() if prefix is not None else None
        return super().iterate_objects(prefix)

    def upload_object(self, file_path: str, object_name: str) -> Object:
        object_name = object_name.lower()
        return super().upload_object(file_path, object_name)

    def upload_object_via_stream(self, iterator: Iterator[bytes], object_name: str) -> Object:
        object_name = object_name.lower()
        return super().upload_object_via_stream(iterator, object_name)


class _BaseAzureStorage(LogMixin):
    def __init__(self,
                 account: str,
                 key: str,
                 container: str,
                 provider: str,
                 host: Optional[str] = None,
                 secure: bool = True,
                 case_sensitive: bool = True) -> None:
        self._account = account
        self._key = key
        self._container = container
        self._provider = getattr(Provider, provider)
        self._host = host or None
        self._secure = secure
        self._case_sensitive = case_sensitive

    @cached_property
    def _driver(self) -> StorageDriver:
        driver = get_driver(self._provider)
        return driver(self._account, self._key, host=self._host, secure=self._secure)

    @cached_property
    def _client(self) -> _Container:
        try:
            container = self._driver.get_container(self._container)
        except ContainerDoesNotExistError:
            try:
                container = self._driver.create_container(self._container)
            except ContainerAlreadyExistsError:
                container = self._driver.get_container(self._container)
        return _Container(container) if self._case_sensitive else _CaseInsensitiveContainer(container)

    @property
    def _generated_suffix(self) -> str:
        return ''

    def access_info(self) -> AccessInfo:
        return AccessInfo(
            account=self._account,
            key=self._key,
            container=self._container,
        )

    def ensure_exists(self):
        # noinspection PyStatementEffect
        self._client

    def delete(self, resource_id: str):
        try:
            resource = self._client.get_object(resource_id)
        except ObjectDoesNotExistError:
            self.log_warning('deleted missing %s', resource_id)
        else:
            resource.delete()
            self.log_debug('deleted %s', resource_id)

    def iter(self, prefix: Optional[str] = None) -> Iterator[str]:
        resources = self._client.iterate_objects(prefix=prefix)

        for resource in resources:
            resource_id = resource.name

            if prefix is not None:
                resource_id = resource_id[len(prefix):]

            if resource_id.endswith(self._generated_suffix):
                resource_id = resource_id[:-len(self._generated_suffix)]

            yield resource_id
            self.log_debug('listed %s', resource_id)


class AzureFileStorage(_BaseAzureStorage):
    def store_file(self, resource_id: str, path: str):
        self.log_debug('storing file %s at %s', path, resource_id)
        self._client.upload_object(path, resource_id)

    def fetch_file(self, resource_id: str) -> str:
        resource = self._client.get_object(resource_id)
        path = create_tempfilename(resource_id)
        resource.download(path)
        self.log_debug('fetched file %s from %s', path, resource_id)
        return path


class _AzureBytesStorage(_BaseAzureStorage):
    _compression = 'gz'

    def store_bytes(self, resource_id: str, content: bytes):
        filename = self._to_filename(resource_id)
        self.log_debug('storing %d bytes at %s', len(content), filename)
        upload = BytesIO()
        upload.write(gzip_bytes(content))
        upload.seek(0)
        self._client.upload_object_via_stream(upload, filename)

    def fetch_bytes(self, resource_id: str) -> bytes:
        filename = self._to_filename(resource_id)
        download = BytesIO()
        resource = self._client.get_object(filename)
        for chunk in resource.as_stream():
            download.write(chunk)
        download.seek(0)
        content = gunzip_bytes(download.read())
        self.log_debug('fetched %d bytes from %s', len(content), filename)
        return content

    def delete(self, resource_id: str):
        filename = self._to_filename(resource_id)
        super().delete(filename)

    def _to_filename(self, resource_id: str) -> str:
        if resource_id.endswith(self._generated_suffix):
            return resource_id
        return f'{resource_id}{self._generated_suffix}'

    @property
    def _generated_suffix(self) -> str:
        return f'.{self._extension}.{self._compression}'

    @property
    def _extension(self) -> str:
        raise NotImplementedError  # pragma: no cover


class AzureTextStorage(_AzureBytesStorage):
    _encoding = 'utf-8'
    _extension = 'txt'

    def store_text(self, resource_id: str, text: str):
        content = text.encode(self._encoding)
        self.store_bytes(resource_id, content)

    def fetch_text(self, resource_id: str) -> str:
        content = self.fetch_bytes(resource_id)
        return content.decode(self._encoding)


class AzureObjectsStorage(LogMixin):
    _compression = 'zstd'
    _compression_level = 20

    def __init__(self, file_storage: AzureFileStorage, resource_id_source: Callable[[], str]):
        self._file_storage = file_storage
        self._resource_id_source = resource_id_source

    def _open_archive_file(self, archive: TarFile, name: str) -> IO[bytes]:
        while True:
            member = archive.next()
            if member is None:
                break
            if member.name == name:
                fobj = archive.extractfile(member)
                if fobj is None:
                    break
                return fobj

        # noinspection PyProtectedMember
        raise ObjectDoesNotExistError(f'File {name} is missing in archive', self._file_storage._driver, archive.name)

    @classmethod
    def _open_archive(cls, path: str, mode: str) -> TarFile:
        extension_index = path.rfind('.')
        if extension_index > -1:
            compression = path[extension_index + 1:]
        else:
            compression = cls._compression

        kwargs = {}
        if compression == 'zstd' and mode == 'w':
            kwargs['level'] = cls._compression_level

        mode = f'{mode}|{compression}'
        return tarfile_open(path, mode, **kwargs)

    def access_info(self) -> AccessInfo:
        return self._file_storage.access_info()

    def ensure_exists(self):
        return self._file_storage.ensure_exists()

    @classmethod
    def compression_formats(cls) -> Iterable[str]:
        return SUPPORTED_FORMATS

    def store_objects(self, upload: Upload, compression: Optional[str] = None) -> Optional[str]:

        compression = compression or self._compression

        resource_id = f'{self._resource_id_source()}.tar.{compression}'

        name, objs, encoder = upload

        num_stored = 0
        with removing(create_tempfilename(resource_id)) as path:
            with self._open_archive(path, 'w') as archive:
                with NamedTemporaryFile() as fobj:
                    num_bytes = 0
                    for obj in objs:
                        encoded = encoder(obj)
                        fobj.write(encoded)
                        num_bytes += len(encoded)
                        num_stored += 1

                    if num_bytes > 0:
                        fobj.seek(0)
                        archive.add(fobj.name, name)

            if num_stored > 0:
                self._file_storage.store_file(resource_id, path)

        self.log_debug('stored %d objects at %s', num_stored, resource_id)
        return resource_id if num_stored > 0 else None

    def fetch_objects(self, resource_id: str, download: Download) -> Iterable[dict]:

        name, decoder = download

        num_fetched = 0
        with removing(self._file_storage.fetch_file(resource_id)) as path:
            with self._open_archive(path, 'r') as archive:
                fobj = self._open_archive_file(archive, name)
                for encoded in fobj:
                    obj = decoder(encoded)
                    if obj is None:
                        continue
                    num_fetched += 1
                    yield obj
        self.log_debug('fetched %d objects from %s', num_fetched, resource_id)

    def delete(self, resource_id: str):
        self._file_storage.delete(resource_id)


class AzureObjectStorage(_AzureBytesStorage):
    _extension = 'msgpack'

    def fetch_object(self, resource_id: str) -> dict:
        serialized = self.fetch_bytes(resource_id)
        return from_msgpack_bytes(serialized)

    def store_object(self, resource_id: str, obj: dict) -> None:
        serialized = to_msgpack_bytes(obj)
        self.store_bytes(resource_id, serialized)



from abc import ABC
from hashlib import sha256
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Tuple
from typing import Union

from libcloud.storage.types import ObjectDoesNotExistError

from opwen_email_server.constants import events
from opwen_email_server.constants import mailbox
from opwen_email_server.constants import sync
from opwen_email_server.services.auth import Auth
from opwen_email_server.services.sendgrid import SendSendgridEmail
from opwen_email_server.services.storage import AzureObjectsStorage
from opwen_email_server.services.storage import AzureObjectStorage
from opwen_email_server.services.storage import AzureTextStorage
from opwen_email_server.utils.email_parser import MimeEmailParser
from opwen_email_server.utils.email_parser import descending_timestamp
from opwen_email_server.utils.email_parser import ensure_has_sent_at
from opwen_email_server.utils.email_parser import get_domain
from opwen_email_server.utils.email_parser import get_domains
from opwen_email_server.utils.email_parser import get_recipients
from opwen_email_server.utils.log import LogMixin
from opwen_email_server.utils.serialization import from_base64
from opwen_email_server.utils.serialization import from_jsonl_bytes
from opwen_email_server.utils.serialization import to_base64
from opwen_email_server.utils.serialization import to_jsonl_bytes
from opwen_email_server.utils.string import is_lowercase
from opwen_email_server.utils.unique import new_email_id

Response = Union[dict, Tuple[str, int]]


class _Action(ABC, LogMixin):
    def __call__(self, *args, **kwargs) -> Response:
        try:
            return self._action(*args, **kwargs)
        except Exception as ex:
            self.log_exception(ex, '%s(args=%r, kwargs=%r)', self.__class__.__name__, args, kwargs)
            raise ex

    def _action(self, *args, **kwargs) -> Response:
        raise NotImplementedError  # pragma: no cover


class Ping(_Action):
    # noinspection PyMethodMayBeStatic
    def _action(self):  # type: ignore
        return 'OK', 200


class SendOutboundEmails(_Action):
    def __init__(self, email_storage: AzureObjectStorage, send_email: SendSendgridEmail):

        self._email_storage = email_storage
        self._send_email = send_email

    def _action(self, resource_id):  # type: ignore
        email = self._email_storage.fetch_object(resource_id)

        success = self._send_email(email)
        if not success:
            return 'error', 500

        self.log_event(events.EMAIL_DELIVERED_FROM_CLIENT, {'domain': get_domain(email.get('from', ''))})  # noqa: E501  # yapf: disable
        return 'OK', 200


class StoreInboundEmails(_Action):
    def __init__(self,
                 raw_email_storage: AzureTextStorage,
                 email_storage: AzureObjectStorage,
                 pending_storage: AzureTextStorage,
                 next_task: Callable[[str], None],
                 email_parser: Callable[[str], dict] = None):

        self._raw_email_storage = raw_email_storage
        self._email_storage = email_storage
        self._pending_storage = pending_storage
        self._next_task = next_task
        self._email_parser = email_parser or MimeEmailParser()

    def _action(self, resource_id):  # type: ignore
        try:
            mime_email = self._raw_email_storage.fetch_text(resource_id)
        except ObjectDoesNotExistError:
            self.log_warning('Inbound email %s does not exist', resource_id)
            return 'skipped', 202

        email = self._email_parser(mime_email)
        email_id = self._store_inbound_email(email)

        self._raw_email_storage.delete(resource_id)
        self._next_task(email_id)

        self.log_event(events.EMAIL_STORED_FOR_CLIENT, {'domain': get_domain(email.get('from') or '')})  # noqa: E501  # yapf: disable
        return 'OK', 200

    def _store_inbound_email(self, email: dict) -> str:
        ensure_has_sent_at(email)
        email_id = new_email_id(email)
        email['_uid'] = email_id

        self._email_storage.store_object(email_id, email)

        for domain in get_domains(email):
            if domain.endswith(mailbox.MAILBOX_DOMAIN):
                self._pending_storage.store_text(f'{domain}/{email_id}', 'pending')

        return email_id


class _IndexEmailForMailbox(_Action):
    def __init__(self, email_storage: AzureObjectStorage, mailbox_storage: AzureTextStorage):
        self._email_storage = email_storage
        self._mailbox_storage = mailbox_storage

    def _action(self, resource_id):  # type: ignore
        email = self._email_storage.fetch_object(resource_id)

        for email_address in self._get_pivot(email):
            domain = get_domain(email_address)
            desc_prefix = descending_timestamp(email['sent_at'])
            if domain.endswith(mailbox.MAILBOX_DOMAIN):
                index = f"{domain}/{email_address}/{self._folder}/{desc_prefix}/{resource_id}"
                self._mailbox_storage.store_text(index, 'indexed')

        self.log_event(events.MAILBOX_EMAIL_INDEXED, {'folder': self._folder})  # noqa: E501  # yapf: disable
        return 'OK', 200

    @property
    def _folder(self) -> str:
        raise NotImplementedError  # pragma: no cover

    def _get_pivot(self, email: dict) -> Iterable[str]:
        raise NotImplementedError  # pragma: no cover


class IndexReceivedEmailForMailbox(_IndexEmailForMailbox):
    _folder = mailbox.RECEIVED_FOLDER

    def _get_pivot(self, email: dict) -> Iterable[str]:
        for email_address in get_recipients(email):
            yield email_address


class IndexSentEmailForMailbox(_IndexEmailForMailbox):
    _folder = mailbox.SENT_FOLDER

    def _get_pivot(self, email: dict) -> Iterable[str]:
        sender = email.get('from')
        if sender:
            yield sender


class StoreWrittenClientEmails(_Action):
    def __init__(self, client_storage: AzureObjectsStorage, email_storage: AzureObjectStorage,
                 user_storage: AzureObjectStorage, next_task: Callable[[str], None]):

        self._client_storage = client_storage
        self._email_storage = email_storage
        self._user_storage = user_storage
        self._next_task = next_task

    def _action(self, resource_id):  # type: ignore
        self._store_emails(resource_id)
        self._store_users(resource_id)
        self._client_storage.delete(resource_id)

        return 'OK', 200

    def _store_emails(self, resource_id):
        emails = self._client_storage.fetch_objects(resource_id, (sync.EMAILS_FILE, from_jsonl_bytes))

        domain = ''
        num_stored = 0
        for email in emails:
            email_id = email['_uid']
            email = self._decode_attachments(email)
            self._email_storage.store_object(email_id, email)

            self._next_task(email_id)

            num_stored += 1
            domain = get_domain(email.get('from', ''))

        self.log_event(events.EMAIL_STORED_FROM_CLIENT, {'domain': domain, 'num_emails': num_stored})  # noqa: E501  # yapf: disable

    def _store_users(self, resource_id):
        users = self._client_storage.fetch_objects(resource_id, (sync.USERS_FILE, from_jsonl_bytes))

        domain = ''
        num_stored = 0
        for user in users:
            email = user['email']
            domain = get_domain(email)
            self._user_storage.store_object(f'{domain}/{email}', user)

            num_stored += 1

        self.log_event(events.USER_STORED_FROM_CLIENT, {'domain': domain, 'num_users': num_stored})  # noqa: E501  # yapf: disable

    @classmethod
    def _decode_attachments(cls, email: dict) -> dict:
        if not email.get('attachments'):
            return email

        for attachment in email['attachments']:
            attachment['content'] = from_base64(attachment['content'])

        return email


class ReceiveInboundEmail(_Action):
    def __init__(self, auth: Auth, raw_email_storage: AzureTextStorage, next_task: Callable[[str], None]):
        self._auth = auth
        self._raw_email_storage = raw_email_storage
        self._next_task = next_task

    def _action(self, client_id=None, email=None, **sendgrid_args):  # type: ignore
        if email is None:
            return 'email cannot be None', 400

        domain = self._auth.domain_for(client_id)
        if not domain:
            self.log_event(events.UNREGISTERED_CLIENT, {'client_id': client_id})  # noqa: E501  # yapf: disable
            return 'client is not registered', 403

        email_id = self._new_email_id(email)

        self._raw_email_storage.store_text(email_id, email)

        self._next_task(email_id)

        self.log_event(events.EMAIL_RECEIVED_FOR_CLIENT, {'domain': domain})  # noqa: E501  # yapf: disable
        return 'received', 200

    @classmethod
    def _new_email_id(cls, email: str) -> str:
        return sha256(email.encode('utf-8')).hexdigest()


class ProcessServiceEmail(_Action):
    def __init__(self,
                 raw_email_storage: AzureTextStorage,
                 email_storage: AzureObjectStorage,
                 next_task: Callable[[str], None],
                 registry: Dict[str, Any],
                 email_parser: Callable[[dict], dict] = None):

        self._raw_email_storage = raw_email_storage
        self._email_storage = email_storage
        self._next_task = next_task
        self._registry = registry
        self._email_parser = email_parser or MimeEmailParser()

    def _action(self, resource_id):  # type: ignore
        try:
            mime_email = self._raw_email_storage.fetch_text(resource_id)
        except ObjectDoesNotExistError:
            self.log_warning('Inbound email %s does not exist', resource_id)
            return 'skipped', 202

        email = self._email_parser(mime_email)

        for address in email.get('to', []):
            try:
                mailer_service = self._registry[address]
            except KeyError:
                self.log_warning('Skipping unknown mailer service: %s', address)
                continue

            formatted_email = mailer_service(email)

            formatted_email_id = new_email_id(formatted_email)
            formatted_email['_uid'] = formatted_email_id

            self._email_storage.store_object(formatted_email_id, formatted_email)

            self._next_task(formatted_email_id)

        self._raw_email_storage.delete(resource_id)
        self.log_event(events.EMAILS_FORMATTED_FOR_CLIENT)  # noqa: E501  # yapf: disable
        return 'OK', 200


class DownloadClientEmails(_Action):
    def __init__(self, auth: Auth, client_storage: AzureObjectsStorage, email_storage: AzureObjectStorage,
                 pending_storage: AzureTextStorage):

        self._auth = auth
        self._client_storage = client_storage
        self._email_storage = email_storage
        self._pending_storage = pending_storage

    def _action(self, client_id, compression):  # type: ignore
        domain = self._auth.domain_for(client_id)
        if not domain:
            self.log_event(events.UNREGISTERED_CLIENT, {'client_id': client_id})  # noqa: E501  # yapf: disable
            return 'client is not registered', 403

        if compression not in self._client_storage.compression_formats():
            self.log_event(events.UNKNOWN_COMPRESSION_FORMAT, {'client_id': client_id})  # noqa: E501  # yapf: disable
            return f'unknown compression format "{compression}"', 400

        delivered = set()

        def mark_delivered(email: dict) -> dict:
            delivered.add(email['_uid'])
            return email

        pending = self._fetch_pending_emails(domain)
        pending = (mark_delivered(email) for email in pending)
        pending = (self._encode_attachments(email) for email in pending)

        resource_id = self._client_storage.store_objects((sync.EMAILS_FILE, pending, to_jsonl_bytes), compression)

        self._mark_emails_as_delivered(domain, delivered)

        self.log_event(events.EMAILS_DELIVERED_TO_CLIENT, {'domain': domain, 'num_emails': len(delivered)})  # noqa: E501  # yapf: disable
        return {
            'resource_id': resource_id,
        }

    def _fetch_pending_emails(self, domain: str) -> Iterable[dict]:
        for email_id in self._pending_storage.iter(f'{domain}/'):
            yield self._email_storage.fetch_object(email_id)

    @classmethod
    def _encode_attachments(cls, email: dict) -> dict:
        if not email.get('attachments'):
            return email

        for attachment in email['attachments']:
            content_bytes = attachment['content']
            attachment['content'] = to_base64(content_bytes)

        return email

    def _mark_emails_as_delivered(self, domain: str, email_ids: Iterable[str]) -> None:
        for email_id in email_ids:
            self._pending_storage.delete(f'{domain}/{email_id}')


class UploadClientEmails(_Action):
    def __init__(self, auth: Auth, next_task: Callable[[str], None]):
        self._auth = auth
        self._next_task = next_task

    def _action(self, client_id, upload_info):  # type: ignore
        domain = self._auth.domain_for(client_id)
        if not domain:
            self.log_event(events.UNREGISTERED_CLIENT, {'client_id': client_id})  # noqa: E501  # yapf: disable
            return 'client is not registered', 403

        resource_id = upload_info['resource_id']

        self._next_task(resource_id)

        self.log_event(events.EMAILS_RECEIVED_FROM_CLIENT, {'domain': domain})  # noqa: E501  # yapf: disable
        return 'uploaded', 200


class RegisterClient(_Action):
    def __init__(self, auth: Auth, client_storage: AzureObjectsStorage, setup_mailbox: Callable[[str, str], None],
                 setup_mx_records: Callable[[str], None], client_id_source: Callable[[], str]):
        self._auth = auth
        self._client_storage = client_storage
        self._setup_mailbox = setup_mailbox
        self._setup_mx_records = setup_mx_records
        self._client_id_source = client_id_source

    def _action(self, domain, owner):  # type: ignore
        client_id = self._client_id_source()

        self._setup_mailbox(client_id, domain)
        self._setup_mx_records(domain)
        self._client_storage.ensure_exists()
        self._auth.insert(client_id, domain, owner)

        self.log_event(events.NEW_CLIENT_REGISTERED, {'domain': domain})  # noqa: E501  # yapf: disable
        return 'OK', 200


class CreateClient(_Action):
    def __init__(self, auth: Auth, task: Callable[[str, str], None]):
        self._auth = auth
        self._task = task

    def _action(self, client, user, **auth_args):  # type: ignore
        domain = client['domain']
        if not is_lowercase(domain):
            return 'domain must be lowercase', 400
        if self._auth.client_id_for(domain) is not None:
            return 'client already exists', 409

        self._task(domain, user)

        self.log_event(events.CLIENT_CREATED, {'domain': domain})  # noqa: E501  # yapf: disable
        return 'accepted', 201


class ListClients(_Action):
    def __init__(self, auth: Auth):
        self._auth = auth

    def _action(self, **auth_args):  # type: ignore
        clients = [{'domain': domain} for domain in self._auth.domains()]

        self.log_event(events.CLIENTS_FETCHED)  # noqa: E501  # yapf: disable
        return {
            'clients': clients,
        }


class GetClient(_Action):
    def __init__(self, auth: Auth, client_storage: AzureObjectsStorage):
        self._auth = auth
        self._client_storage = client_storage

    def _action(self, domain, user, **auth_args):  # type: ignore
        if not is_lowercase(domain):
            return 'domain must be lowercase', 400

        client_id = self._auth.client_id_for(domain)
        if client_id is None:
            return 'client does not exist', 404

        if not self._auth.is_owner(domain, user):
            return 'client does not belong to the user', 403

        access_info = self._client_storage.access_info()

        self.log_event(events.CLIENT_FETCHED, {'domain': domain})  # noqa: E501  # yapf: disable
        return {
            'client_id': client_id,
            'storage_account': access_info.account,
            'storage_key': access_info.key,
            'resource_container': access_info.container,
        }


class DeleteClient(_Action):
    def __init__(self, auth: Auth, delete_mailbox: Callable[[str, str], None], delete_mx_records: Callable[[str], None],
                 mailbox_storage: AzureTextStorage, pending_storage: AzureTextStorage,
                 user_storage: AzureObjectStorage):
        self._auth = auth
        self._delete_mailbox = delete_mailbox
        self._delete_mx_records = delete_mx_records
        self._mailbox_storage = mailbox_storage
        self._pending_storage = pending_storage
        self._user_storage = user_storage

    def _action(self, domain, user, **auth_args):  # type: ignore
        if not is_lowercase(domain):
            return 'domain must be lowercase', 400

        client_id = self._auth.client_id_for(domain)
        if client_id is None:
            return 'client does not exist', 404

        if not self._auth.is_owner(domain, user):
            return 'client does not belong to the user', 403

        self._delete_mailbox(client_id, domain)
        self._delete_mx_records(domain)
        self._delete_index(self._pending_storage, domain)
        self._delete_index(self._mailbox_storage, domain)
        self._delete_index(self._user_storage, domain)
        self._auth.delete(client_id, domain)

        self.log_event(events.CLIENT_DELETED, {'domain': domain})  # noqa: E501  # yapf: disable
        return 'OK', 200

    @classmethod
    def _delete_index(cls, storage: Union[AzureTextStorage, AzureObjectStorage], domain: str) -> None:
        for prefix in storage.iter(f'{domain}/'):
            storage.delete(f'{domain}/{prefix}')


class CalculateNumberOfUsersMetric(_Action):
    def __init__(self, auth: Auth, user_storage: AzureObjectStorage):
        self._auth = auth
        self._user_storage = user_storage

    def _action(self, domain, user, **auth_args):  # type: ignore
        if not self._auth.is_owner(domain, user):
            return 'client does not belong to the user', 403

        users = sum(1 for _ in self._user_storage.iter(f'{domain}/'))

        return {
            'users': users,
        }


class CalculatePendingEmailsMetric(_Action):
    def __init__(self, auth: Auth, pending_storage: AzureTextStorage):
        self._auth = auth
        self._pending_storage = pending_storage

    def _action(self, domain, user, **auth_args):  # type: ignore
        if not self._auth.is_owner(domain, user):
            return 'client does not belong to the user', 403

        pending_emails = sum(1 for _ in self._pending_storage.iter(f'{domain}/'))

        return {
            'pending_emails': pending_emails,
        }



from socket import create_connection
from socket import gethostbyname


def check_connection(hostname: str, port: int) -> bool:
    try:
        host = gethostbyname(hostname)
        with create_connection((host, port)):
            return True
    except OSError:
        pass
    return False
